from typing import Optional, TextIO, Tuple

from cocluremig.utils.gitutils import normalize_id_name


class BaseDotWriter:
    """
    Base class for dot-File-Writing
    """

    def __init__(self, file_writer: TextIO):
        """
        Creates the dot writer

        @param file_writer: a file-writer to write the dot-file into
        """
        self._file_writer = file_writer
        self.__closed = False

    def write_head(self, name: str):
        """
        writes header

        @param name: graph name
        """
        self._file_writer.write(" digraph ")
        self._file_writer.write(normalize_id_name(name))
        self._file_writer.write(" {")

    def write_edge(self, edge: Tuple[str, str]):
        """
        writes an edge

        @param edge: edge to write
        """
        (v1, v2) = edge
        self._file_writer.write("\"")
        self._file_writer.write(v1)
        self._file_writer.write("\" -> \"")
        self._file_writer.write(v2)
        self._file_writer.write("\"; ")

    def write_vertex(self, vertex: str, label: Optional[str] = None, color: Optional[str] = None):
        """
        writes a vertex to the graph

        @param vertex: the vertex to write
        @param label: a label to use instead of identifier
        @param color: a color for the vertex
        """
        self._file_writer.write("\"")
        self._file_writer.write(vertex)
        self._file_writer.write("\"")
        if [x for x in [label, color] if x]:
            self._file_writer.write(" [")
            if label:
                self._file_writer.write("label=\"")
                self._file_writer.write(label)
                self._file_writer.write("\"")
            if label and color:
                self._file_writer.write(",")
            if color:
                self._file_writer.write("color=\"")
                self._file_writer.write(color)
                self._file_writer.write("\"")
            self._file_writer.write("]")
        self._file_writer.write("; ")

    def write_footer(self):
        """
        writes footer
        """
        self._file_writer.write("}")

    def close(self):
        """
        closes the file
        """
        self._file_writer.flush()
        self.__closed = True

    def __del__(self):
        if not self.__closed:
            self.close()
        self._file_writer.close()



"""
clustering algorithm
"""
import logging
import math
from builtins import isinstance
from collections import OrderedDict
from numbers import Real
from typing import Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import git
import pandas as pd

from cocluremig.analyzer import graph_analyzer
from cocluremig.analyzer.graph_analyzer import AdjencyMatrix, Edge, Graph, Node


def create_meta_graph_from_matrix(graph: Graph) -> AdjencyMatrix:
    """
    Converts the given graph into a Meta_graph
    ( https://doi.org/10.1109/ICSME.2014.48 )

    @param graph: The input graph
    @return: the meat-graph as adjency matrix
    """
    # https://doi.org/10.1109/ICSME.2014.48
    adjency_matrix: AdjencyMatrix
    if not isinstance(graph, AdjencyMatrix):
        # transform to adjency_matrix
        adjency_matrix = graph_analyzer.edge_list_to_directed_adjency_matrix(graph)
    else:
        adjency_matrix = graph
    logging.info("Creating meta graph")
    m_nodes = graph_analyzer.get_merge_or_branch_nodes(adjency_matrix)
    for vertex in adjency_matrix.index.tolist():
        logging.debug("Processing %s", vertex)
        if vertex in m_nodes:
            continue
        # with only one incoming and one outgoing edge per node,
        # it can be replaces by a new edge incoming -> outgoing
        # index row != 0 -> index column != 0
        ins = adjency_matrix.loc[vertex, :]
        outs = adjency_matrix.loc[:, vertex]
        in_index = ins[ins != 0].index.tolist()
        out_index = outs[outs != 0].index.tolist()
        if len(out_index) != 1 or len(in_index) != 1:
            # continue
            raise ValueError(vertex, in_index, out_index, graph)
        out_index = out_index[0]
        in_index = in_index[0]
        adjency_matrix.at[out_index, in_index] = adjency_matrix.at[out_index, in_index] + \
                                                 ins[in_index] + outs[out_index]
        if adjency_matrix.at[out_index, in_index] < 1:
            raise ValueError(vertex, in_index, out_index, graph)
        adjency_matrix = adjency_matrix.drop(vertex, axis='columns')
        adjency_matrix = adjency_matrix.drop(vertex, axis='rows')
    return adjency_matrix


def create_meta_graph(edge_list: List[str]) -> Iterable[Tuple[str, str, Optional[List[str]]]]:
    """
     Converts the given edge-list of a graph graph into a Meta_graph
    ( https://doi.org/10.1109/ICSME.2014.48 )
    Each meta-edge is given the list sequential commits
    composing that edge as metadata
    @param edge_list: an list of edges
    @return: the Meta-Graph as Edge list with metadata
    """
    # https://doi.org/10.1109/ICSME.2014.48
    db_con = graph_analyzer.edge_list_to_memory_db(edge_list)
    sequentials = graph_analyzer.get_sequential_commits(db_con)
    seq_paths = graph_analyzer.get_sequential_paths(db_con)
    edges = []
    for (v1, v2) in edge_list:
        if not (v1 in sequentials or v2 in sequentials):
            edges.append((v1, v2, []))
    for path in seq_paths:
        edges.append((path[0], path[-1], path[1:-1]))
    return edges


def create_quasi_meta_graph(edges: List[Edge]) \
        -> Tuple[Iterable[Edge], Dict[int, List[str]]]:
    """
    Creates a Quasi-Meta-Graph from given edge-list.
    this is a derivation of the meta-graph ( https://doi.org/10.1109/ICSME.2014.48 ) ,
    where every crossroad node is connected to up to one itermediate node and
    merging is done on nodes not edges
    @param edges: an edge-list of a graph
    @return: A tuple of
        the edge-list of the meta-graph and
        a dictionary mapping clusters to their contained nodes
    """
    db_con = graph_analyzer.edge_list_to_memory_db(edges)
    crossroads = graph_analyzer.get_merge_or_branch_nodes(db_con)
    weight_fuction = lambda x, y, repo: 0 if (x in crossroads or y in crossroads) else 1
    (quasi_meta, values) = cluster(edges, None, weight_fuction, 0, 1)
    values = [[v1, v2, res] for (v1, v2), res in values.items()]
    (graph, clusters, replacement_table) = make_visuable_clustered_graph_step(quasi_meta)
    return (graph, clusters)


DataSource = Union[git.Repo, pd.DataFrame]


def cluster(topology: List[Edge], data: DataSource,
            heuristic: Callable[
                [Node, Node, DataSource], Real],
            steps: Union[Real, List[Real]] = 0,
            threshold: Real = 0) \
        -> Tuple[Iterable[Iterable[Edge]], Dict[Edge, Real]]:
    """
    Performs topology-aware-clustering on the given topology
    @param topology: an edge-list
    @param data: an additional data-source for our heurisitc
    @param heuristic: a function which gives all edges a weight
    @param steps: Either a List of predifined values of values, which determine the clustering steps
                    or a value to define merge step size,
                    when left empty every value of the heuristic is its own step
    @param threshold: a value, which when reached stops merging
    @return: A tuple of
        1. A List of Lists where every sublist represtens the edges merged in on step
        2. the mapping of edges to their heuristic
    """
    values: OrderedDict[Edge, Real] = \
        OrderedDict(sorted(
            [((e1, e2), heuristic(e1, e2, data))
             for (e1, e2) in topology],
            key=lambda x: x[1], reverse=True))
    curr_val: Real = next(iter(values.items()))[1]
    if steps:
        if isinstance(steps, Real):
            step_size = steps
            back_iter = iter(reversed(values.items()))
            fore_iter = iter(values.items())
            max_val = next(fore_iter)[1]
            min_val = next(back_iter)[1]
            while math.isinf(min_val):
                min_val = next(back_iter)[1]
            while math.isinf(max_val):
                max_val = next(fore_iter)[1]
            max_val = math.ceil(max_val / step_size)
            min_val = math.ceil(min_val / step_size)
            step_no = max(abs(max_val), abs(min_val))
            max_val *= step_size
            min_val *= step_size
            steps = [c * step_size for c in range(0, step_no)]
            if not steps:
                steps = [step_size, step_size / 2]
        if isinstance(steps, list):
            if curr_val <= 0 and steps[1] >= 0:
                steps = list(map(lambda x: -x, steps))
            steps = list(sorted(steps, reverse=True))
    steps: List[Real]
    merged_this_step: List[Edge] = []
    unmerged: Set[Edge] = set(topology)
    out: Iterable[Iterable[Edge]] = []
    curr_step = 1
    for (edge, value) in values.items():
        if steps and value <= steps[curr_step]:
            curr_step += 1
            if merged_this_step:
                out.append(list(merged_this_step))
                merged_this_step = []
            if curr_step == len(steps):
                break
        elif not steps and value < curr_val:
            if merged_this_step:
                out.append(list(merged_this_step))
                merged_this_step = []
            curr_val = value
        if threshold and value < threshold:
            break

        # merge vertices
        merged_this_step.append(edge)
        unmerged.remove(edge)
    # only add if non-empty
    if merged_this_step:
        out.append(merged_this_step)
    if unmerged:
        out.append(list(unmerged))
    return (out, values)


def make_visuable_clustered_graph_step(
        merge_steps: Iterable[Iterable[Edge]],
        step: int = 0,
        replacement_table: Optional[Dict[str, Set[str]]] = None) \
        -> Tuple[Iterable[Edge],
                 Dict[int, List[str]],
                 Dict[str, Set[str]]]:
    """
    Converts the given output of our topology-aware-clustering into a usable edge-list
    @param merge_steps:  A List of Lists where every sublist represtens the edges merged in on step
    @param step: which step to currently merge
    @param replacement_table: A mapping of nodes to their merged clusters (default: empty)
    @return: A tuple of
            1. an edge list of the output graph
            2. a dictionary mapping the cluster-no to the contained shas
            3. a mapping of nodes to their clusters (to insert into the following merge-step)
    """
    if replacement_table is None:
        replacement_table = {}
    merged_edges: Set[Tuple[Edge]] = set()
    # populate replacement table
    for (vertex_1, vertex_2) in merge_steps[step]:
        c_list_1 = replacement_table.get(vertex_1, {vertex_1})
        c_list_2 = replacement_table.get(vertex_2, {vertex_2})
        c_list_1 = c_list_1.union(c_list_2)
        for vertex in c_list_1:
            replacement_table[vertex] = c_list_1
    # name clusters
    clusters: Dict[int, List[str]] = {}
    for vertex in replacement_table.values():
        clusters[id(vertex)] = list(vertex)
    for unmerged in merge_steps[(step + 1):]:
        for (vertex_1, vertex_2) in unmerged:
            new_vertex_1 = replacement_table.get(vertex_1)
            if new_vertex_1:
                new_vertex_1 = "c_" + str(id(new_vertex_1))
            else:
                new_vertex_1 = vertex_1
            new_vertex_2 = replacement_table.get(vertex_2)
            if new_vertex_2:
                new_vertex_2 = "c_" + str(id(new_vertex_2))
            else:
                new_vertex_2 = vertex_2
            # remove self-referencing edges
            if new_vertex_1 != new_vertex_2:
                merged_edges.add((new_vertex_1, new_vertex_2))
    return (merged_edges, clusters, replacement_table)


def visuable_cluster_blind(edges: Iterable[Edge], replacement_table: Dict[str, Set[str]]) \
        -> Iterable[Edge]:
    """
    Creates a new graph based on given repalcement information
    @param edges:  an edge list
    @param replacement_table: a mapping from a vertex to a cluster/set of vertices
    @return: the edge list of the output-graph
    """
    new_edges, clusters, replacement_out = \
        make_visuable_clustered_graph_step(
            [[], list(edges)], 0, replacement_table)
    return new_edges



"""
Graph visualization modules
"""
from typing import List

"""DOT_COLORS: List[str] = [c.strip()
                         for c in open(
        os.path.dirname(
            os.path.abspath(__file__)) +
        os.path.sep + "colors_short.csv")]"""
DOT_COLORS: List[str] = ['aliceblue', 'antiquewhite', 'aqua',
                         'aquamarine', 'azure', 'beige',
                         'bisque', 'black', 'blanchedalmond',
                         'blue', 'blueviolet', 'brown',
                         'burlywood', 'cadetblue', 'chartreuse',
                         'chocolate', 'coral', 'cornflowerblue',
                         'cornsilk', 'crimson', 'cyan', 'darkblue',
                         'darkcyan', 'darkgoldenrod', 'darkgray',
                         'darkgreen', 'darkgrey', 'darkkhaki',
                         'darkmagenta', 'darkolivegreen', 'darkorange',
                         'darkorchid', 'darkred', 'darksalmon',
                         'darkseagreen', 'darkslateblue', 'darkslategray',
                         'darkslategrey', 'darkturquoise', 'darkviolet',
                         'deeppink', 'deepskyblue', 'dimgray',
                         'dimgrey', 'dodgerblue', 'firebrick',
                         'floralwhite', 'forestgreen', 'fuchsia',
                         'gainsboro', 'ghostwhite', 'gold',
                         'goldenrod', 'green', 'greenyellow',
                         'grey', 'honeydew', 'hotpink',
                         'indianred', 'indigo', 'invis',
                         'ivory', 'khaki', 'lavender',
                         'lavenderblush', 'lawngreen',
                         'lemonchiffon', 'lightblue', 'lightcoral',
                         'lightcyan', 'lightgoldenrod', 'lightgoldenrodyellow',
                         'lightgray', 'lightgreen', 'lightgrey',
                         'lightpink', 'lightsalmon', 'lightseagreen',
                         'lightskyblue', 'lightslateblue', 'lightslategray',
                         'lightslategrey', 'lightsteelblue', 'lightyellow',
                         'lime', 'limegreen', 'linen',
                         'magenta', 'maroon', 'mediumaquamarine',
                         'mediumblue', 'mediumorchid',
                         'mediumpurple', 'mediumseagreen', 'mediumslateblue',
                         'mediumspringgreen', 'mediumturquoise',
                         'mediumvioletred', 'midnightblue', 'mintcream',
                         'mistyrose', 'moccasin', 'navajowhite', 'navy',
                         'navyblue', 'none', 'oldlace', 'olive',
                         'olivedrab', 'orange', 'orangered',
                         'orchid', 'palegoldenrod', 'palegreen',
                         'paleturquoise', 'palevioletred', 'papayawhip',
                         'peachpuff', 'peru', 'pink', 'plum',
                         'powderblue', 'purple', 'rebeccapurple',
                         'red', 'rosybrown', 'royalblue', 'saddlebrown',
                         'salmon', 'sandybrown', 'seagreen', 'seashell',
                         'sienna', 'silver', 'skyblue',
                         'slateblue', 'slategray',
                         'slategrey', 'snow', 'springgreen',
                         'steelblue', 'tan', 'teal',
                         'thistle', 'tomato', 'transparent',
                         'turquoise', 'violet', 'violetred',
                         'webgray', 'webgreen', 'webgrey',
                         'webmaroon', 'webpurple', 'wheat',
                         'white', 'whitesmoke',
                         'yellow', 'yellowgreen']
"""
All available dot base colors
"""



"""
Base classes for commit analytics
"""
import functools
import logging
from typing import Callable, cast, List, Optional, TypeVar, Union

import git

from cocluremig.utils import tweaks

T = TypeVar('T')
"""Single Blob metric result"""

V = TypeVar('V')
"""Aggregated repo metric"""


class RepoFileMetricAnalyzer:
    """
        Analyses the given Repo for a given metric

        On consecuting commit-calls works incrementally
        by caching and working on the commits' diffs
    """

    def __init__(self,
                 repo: git.Repo,
                 metric_func: Callable[[git.Object], T],
                 aggregator_func: Optional[Callable[[Union[T, V], Union[T, V]], V]] = None,
                 aggregator_init: Optional[Union[T, V]] = None,
                 path_filter: Optional[Callable[[str], bool]] = None):
        """
        Initalizes the analyzer

        @param repo: the git repository to work on
        @param metric_func:
                a metric to retrieve from the repository's files
        @param aggregator_func:
                a function to aggregate the metric values
                (default: list of results creation)
        @param aggregator_init:
                the neutral element of the aggregation function
        @param path_filter: path/file skipping filter
        """
        logging.debug("Creating Analyzer")
        self.__repo = repo
        self.__metric_func = metric_func
        if aggregator_func:
            self.__aggregator_func = aggregator_func
            self.__aggregator_init = aggregator_init
        # if no aggregation function is given use list concatenation
        else:
            self.__aggregator_func = lambda x, y: x + y if isinstance(y, list) else x + [y]
            self.__aggregator_init = []
        self.__path_filter = path_filter
        # per instance cache
        cache_size = tweaks.get_cache_size()
        logging.debug("Cache size: %i", cache_size)
        self._apply_metric = \
            functools.lru_cache(maxsize=cache_size)(self._apply_metric)  # type: ignore

    def apply_metric(self, sha: Union[str, git.Object]) -> V:
        """
        calculates the metric for given commit

        @param sha: the commit to analyze
        @return: metric
        """
        if isinstance(sha, str):
            sha = self.__repo.commit(sha)
        if sha.type == "tag":
            # recursive call because tag-object can point to other tag-object
            return self.apply_metric(sha.object)
        if sha.type == "commit":
            logging.debug("analyzing commit %s", sha.hexsha)
            return self._apply_metric(sha.tree)
        raise ValueError("Illegal Object Type", (self.__repo, sha))

    def _apply_metric(self, sha: git.Object) -> Union[T, V]:
        logging.log(logging.DEBUG - 5, "Processing %s", (sha.type, sha.hexsha))
        try:
            if sha.type == "blob":
                return self.__metric_func(sha)
            if sha.type == "tree":
                cast(git.Tree, sha)
                if self.__path_filter:
                    if not self.__path_filter(sha.path):
                        return self.__aggregator_init
                # Error in gitPython documentation so iterate over blobs and trees seperately
                vals: List[Union[T, V]] = [self._apply_metric(gitobject) for gitobject in sha.blobs]
                vals += [self._apply_metric(gitobject) for gitobject in sha.trees]
                if None in vals:
                    logging.warning("None Value detected: %i", vals.index(None))
                return functools.reduce(self.__aggregator_func, vals, self.__aggregator_init)
            # somehow sha can also become a submodule, ignore those
            raise ValueError("Illegal Object Type", (self.__repo, sha.path, sha.hexsha, sha.type))
        except RecursionError as error:
            raise ValueError((self.__repo, sha)) from error



"""
Git repo helper functions
"""
import functools
import os
import re
import tempfile
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Tuple
from urllib.error import URLError

import git

from cocluremig.utils import github


def normalize_id_name(name: str) -> str:
    """
    Normalizes Names by removing special chars and whitespaces
    @param name: a string to clean
    @return: a cleaned string
    """
    name = re.sub(r"[^\w]", " ", name.casefold()).strip()  # Remove special Chars
    return "_".join(filter(lambda x: x, map(lambda x: x.strip().capitalize(),
                                            name.split(
                                                " "))))
    # Replace whitespace with Underscope and Capitalize Words


def get_remote_name(url: str) -> str:
    """
    Shortens remote identifiers
    @param url: the remotes url
    @return: an id
    """
    h = get_host(url)
    if "github" in h:
        user, project = github.parse_url(url)
        return normalize_id_name("GH___" + user + "__" + project)
    return normalize_id_name(url)


def get_repo(url: str, directory=None, remotes: Optional[List[str]] = None, add_all_remotes=False,
             no_fetch=False) -> git.Repo:
    """
    Clones or gets a git repo (bare)
    @param url: the repo's url
    @param directory: the repo save directory
    @param remotes: a list of remote urls
    @param add_all_remotes:
        if set to true tries to get & add all remotes
        via git-provider-api (only github implemented yet)
    @param no_fetch:
        if set to true does not fetch data from other machines
    @return:
        a git repo object
    """
    fldr_name = ".".join(url.split("/")[-2:])
    if directory is None:
        fldr_name = tempfile.gettempdir() + os.path.sep + fldr_name
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        fldr_name = directory + os.path.sep + fldr_name
    if no_fetch:
        return git.Repo(fldr_name)
    if add_all_remotes:
        main_host = get_host(url)
        if main_host == "github.com":
            (user, project) = github.parse_url(url)
            if remotes is None:
                remotes = []
            try:
                (gh_remotes, _) = github.find_forks(user, project)
                remotes += gh_remotes
            except URLError:
                pass
    if os.path.exists(fldr_name):
        os.putenv("GIT_TERMINAL_PROMPT", "0")  # disable credential input prompts
        repo = git.Repo(fldr_name)
        if remotes is not None and not len(repo.remotes) == len(remotes) + 1:
            remote_names = list(map(lambda x: x.name, repo.remotes))
            for r in remotes:
                l_name = get_remote_name(r)
                if l_name not in remote_names:
                    nr = repo.create_remote(l_name, r)
                    with nr.config_writer as config_writer:
                        config_writer.set("fetch", "+refs/heads/*:refs/remotes/" + l_name + "/*")
    else:
        repo = git.Repo.clone_from(url, fldr_name, bare=True)
        if remotes is not None:
            for remote in remotes:
                repo.create_remote(get_remote_name(remote), remote)
        for remote in repo.remotes:
            with remote.config_writer as config_writer:
                config_writer.set("fetch", "+refs/heads/*:refs/remotes/" + remote.name + "/*")
        with repo.config_writer() as config_writer:
            config_writer.set_value("core", "askpass", "")
    try:
        for remote in repo.remotes:
            remote.fetch()
    except git.exc.GitCommandError as e:
        pass
    return repo


def get_host(url: str) -> str:
    """
    retrieves host-name from url
    @param url: a url
    @return: the host name
    """
    if url.startswith("http"):
        url = re.sub(r"http(s)?://", "", url)
        return url.split("/")[0]
    if url.startswith("ssh://"):
        url = url.lstrip("ssh://")
        return url.split(":")[0].split("@")[-1]
    if url.startswith("git://"):
        url = url.lstrip("git://")
        return url.split("/")[0]
    # assuming scp-style-ssh
    return url.split(":")[0].split("@")[-1]


def get_ref_names(repo: git.Repo,
                  ignore_tags: bool = False,
                  ignore_remotes: bool = False,
                  ignore_remotes_if_local_named: bool = True) \
        -> Dict[str, List[str]]:
    """
    Gets reference (branch, tag) names from repo

    @param repo: a git repository
    @param ignore_tags:
        if set to true does not consider tags
    @param ignore_remotes:
        if set to true does not consider names from remote repos
    @param ignore_remotes_if_local_named:
        if set to true does to consider names from remote repos
            if a name is alread present locally
    @return: a commit sha to refernce name(s) dictionary
    """
    refs: Dict[str, List[str]] = defaultdict(list)
    if not ignore_tags:
        for tag in repo.tags:
            refs[tag.commit.hexsha].append(tag.name)
    for local_ref in repo.heads:
        refs[local_ref.commit.hexsha].append(local_ref.name)
    if not ignore_remotes:
        for remote in repo.remotes:
            for remote_ref in remote.refs:
                if not refs.get(remote_ref.object.hexsha) or not ignore_remotes_if_local_named:
                    refs[remote_ref.object.hexsha].append(remote_ref.name)
    return refs


@functools.lru_cache(maxsize=2)
def get_edge_list(repo: git.Repo) -> Tuple[Iterable[Tuple[str, str]], Iterable[git.Commit]]:
    """
    Gets edge list of commit-graph
    @param repo:
    @return: A Tuple of
        - the edge list
        - the set of commits
    """
    edges: Iterable[Tuple[str, str]] = set()
    refs: Dict[str, List[str]] = get_ref_names(repo)
    processed_commits = set()
    for starting_point in refs.keys():
        for commit in repo.iter_commits(starting_point):
            sha = commit.hexsha
            for p in commit.parents:
                edges.add((sha, p.hexsha))
            processed_commits.add(commit)
    return (edges, processed_commits)



"""
Github-specific data-retrieval
"""
import json
import logging
import os
import re
import typing
from http.client import HTTPResponse
from typing import Pattern
from urllib.parse import urlencode
from urllib.request import Request, urlopen

GITHUB_PREFIX: Pattern[str] = re.compile(r"^(((((git)|(ssh)|(http(s)?))://)?(git@)?github\.com(/|:)))")
GIT_POSTFIX: Pattern[str] = re.compile(r"\.git$")


def parse_url(url: str) -> typing.Tuple[str, str]:
    """
    retrieves project identifier from url
    @param url: a github url
    @return: user, project tuple
    """
    url = GITHUB_PREFIX.sub("", url.strip())
    url = GIT_POSTFIX.sub("", url)
    r_info = url.split("/")
    # user , project
    return (r_info[0], r_info[1])


def find_forks(user: str, project: str, params: typing.Dict = {'per_page': 100}, rec_forks: bool = True) \
        -> typing.Tuple[typing.List[str], typing.List[typing.Dict[str, typing.Any]]]:
    """
    Gets forks for github project
    @param user: project owner name
    @param project: project name
    @param params: github api-params
    @param rec_forks: if set to true also retrieves forks of forks
    @return: A tuple of
        - the fork urls
        - full github api-json for projects
    """
    json_out = []
    GITHUB_FORK_URL = "https://api.github.com/repos/{user}/{project}/forks"
    url = GITHUB_FORK_URL.format(user=user, project=project)
    if params:
        url = url + "?" + urlencode(params)
    header: __Header
    with _open_github_url(url) as result:
        result: HTTPResponse
        header = __parse_header(result)
        json_out += json.loads(result.read().decode(header['encoding']))
    while header.get("links"):
        if header.get("links").get("next"):
            with _open_github_url(header.get("links").get("next")) as next_page:
                header = __parse_header(next_page)
                json_out += json.loads(next_page.read().decode(header['encoding']))
        else:
            break
    if rec_forks:
        forked_forks = filter(lambda x: x['forks_count'], json_out)
        raw_ff = []
        for fork in forked_forks:
            _, raw = find_forks(fork['owner']['login'], fork['name'])
        json_out += raw_ff
    return (list(map(lambda x: x['clone_url'], filter(lambda x: not x['private'], json_out))), json_out)


def _open_github_url(url: str) -> HTTPResponse:
    logging.debug("opening %s", url)
    request = Request(url)
    if os.environ.get("GITHUB_TOKEN"):
        request.add_header("Authorization", "token " + os.environ.get("GITHUB_TOKEN"))
    else:
        logging.warning("No Github OAUTH-Token provided. Using public API.")
    return urlopen(request)


class __Header(typing.TypedDict):
    content: str
    encoding: str
    links: typing.Dict[str, str]


def __parse_header(response: HTTPResponse) -> __Header:
    headers = {'content':
                   response.getheader("content-type").split(";")[0],
               'encoding':
                   response.getheader("content-type").split(";")[-1].split("=")[-1].strip(),
               'links': {}}
    links = response.getheader("link")
    if links:
        for l in links.split(","):
            rel = ""
            url = ""
            for t in l.split(";"):
                if "rel=\"" in t:
                    rel = t.strip().lstrip("rel=").strip("\"")
                else:
                    url = t.strip().lstrip("<").rstrip(">")
            headers['links'][rel] = url
    return headers



"""
Custom Containers for Metric results
"""
from numbers import Real
from typing import NamedTuple


class AverageElement(NamedTuple):
    """
    Represents an element of an averageable Value
    Enables calculating incremental average (mean) values
    """
    sum: Real
    no_of_elements: int = 1

    def __add__(self, other):
        if isinstance(other, AverageElement):
            return AverageElement(self.sum + other.sum, self.no_of_elements + other.no_of_elements)
        if not other:
            return self
        else:
            raise TypeError

    def avg(self) -> float:
        """
        calculates the concrete value

        @return: the concrete float value of the average
        """
        return self.sum * 1.0 / self.no_of_elements

    def __float__(self):
        return self.avg()


AVERAGE_NEUTRAL: AverageElement = AverageElement(0, 0)
"""Neutral element of average element addition"""



"""
Concrete Commit analytics
"""
from collections import Counter
from typing import Dict, Iterable, Set, Union

import git

from cocluremig.analyzer.commit import aggregators, blob_inspectors
from cocluremig.analyzer.commit.base_analyzer import RepoFileMetricAnalyzer


def get_basic_commit_data(commit: git.Commit) -> Dict[str, Union[int, str, bool]]:
    """
    gets all data directly stored in commits

    @param commit: a git commit
    @return: a dictionary of commit info
    """
    data = dict(sha=commit.hexsha,
                date_committed=commit.committed_date,
                date_authored=commit.authored_date,
                signed=bool(commit.gpgsig),
                author_name=commit.author.name,
                author_mail=commit.author.email,
                committer_name=commit.committer.name,
                committer_mail=commit.committer.email)
    return data


def get_file_number_per_extension_analyzer(repo: git.Repo) -> RepoFileMetricAnalyzer:
    """
    Gets an analyzer for file number per extension (which returns a dictionary)

    @param repo: a git repository
    @return: the analyzer-object
    """
    return RepoFileMetricAnalyzer(repo,
                                  lambda x: {blob_inspectors.get_file_extension(x): int(1)},
                                  aggregators.additive_dict_merge, {})


def get_file_number_analyzer(repo: git.Repo) -> RepoFileMetricAnalyzer:
    """
    Gets an analyzer for file number (which returns an integer)
    @param repo: a git repository
    @return: the analyzer-object
    """
    return RepoFileMetricAnalyzer(repo,
                                  lambda x: 1,
                                  lambda x, y: x + y,
                                  0)


def get_lizard_analyzer(repo: git.Repo) -> RepoFileMetricAnalyzer:
    """
    Gets an analyzer for lizard (http://lizard.ws) metrics (a list of lizard results)

    @param repo: a git repository
    @return: the analyzer-object
    """
    return RepoFileMetricAnalyzer(repo,
                                  blob_inspectors.get_code_metrics_lizard)


def get_lizard_analyzer_avg(repo: git.Repo) -> RepoFileMetricAnalyzer:
    """
    Gets an analyzer for lizard metrics on average (which returns a dictionary of average elements)

    @param repo: a git repository
    @return: the analyzer-object
    """
    return RepoFileMetricAnalyzer(repo,
                                  blob_inspectors.get_lizard_metrics_avarageable,
                                  aggregators.additive_dict_merge,
                                  {})


def get_lines_of_text_analyzer(repo: git.Repo) -> RepoFileMetricAnalyzer:
    """
    Gets an analyzer for lines of text (which returns an int)

    @param repo: a git repository
    @return: the analyzer-object
    """
    return RepoFileMetricAnalyzer(repo,
                                  blob_inspectors.get_lines_of_text,
                                  lambda x, y: x + y,
                                  0)


def get_file_text_size_analyzer(repo: git.Repo) -> RepoFileMetricAnalyzer:
    """
    Gets the text or file size per file-type
    lines of text for text-based files and file size for binary files

    @param repo: a git repo
    @return: the analyzer-object
    """
    return RepoFileMetricAnalyzer(repo,
                                  blob_inspectors._get_mixed_size_per_extension,
                                  aggregators.additive_dict_merge,
                                  {})


def get_files(sha: git.objects.Commit) -> Set[str]:
    """
    gets all file-paths available in given commit

    @param sha: the commit
    @return: a list of paths
    """
    return {blob.path for blob in sha.tree.traverse() if blob.type == "blob"}


def get_all_commits(repo: git.Repo) -> Iterable[git.objects.Commit]:
    """
    Gets all commits for given repo

    @param repo: a git repository
    @return: the set of all commits
    """
    return {commit for starting_point in repo.heads
            for commit in repo.iter_commits(starting_point.commit)}


def get_dominating_language(repo: git.Repo) -> str:
    """
    Gets dominating (most commonly used) language for repository across all commits

    @param repo: a git repository
    @return: the file extension
    """
    extension_getter = lambda x: x.split(".")[-1]
    ext_cnt = Counter([extension_getter(f)
                       for commit in get_all_commits(repo)
                       for f in get_files(commit)])
    return ext_cnt.most_common(1)[0][0]



"""
    Branch recovery method
    from Michaud, Guarnera, Collard & Maletic's method
    https://doi.org/10.1109/ICSME.2016.39
"""

import logging
import sqlite3
from collections import Counter
from typing import *

import git

# https://doi.org/10.1109/ICSME.2016.39
from cocluremig.analyzer.graph_analyzer import Edge, edge_list_to_memory_db, get_merge_or_branch_nodes
from cocluremig.utils.gitutils import get_edge_list, get_ref_names


def recover_branches(repo: git.Repo) -> Tuple[Dict[str, str], Iterable[Edge]]:
    """
    Associates commits with their branch of origin
    according to Michaud, Guarnera, Collard & Maletic's method
    https://doi.org/10.1109/ICSME.2016.39

    @param repo: a Git repository
    @return: a Tuple of
        a dictionary mapping commits to their branches
        and
        the edge list of the commit graph
    """
    (edges, _) = get_edge_list(repo)
    db: sqlite3.Connection = edge_list_to_memory_db(edges)
    refs: Dict[str, List[str]] = get_ref_names(repo, True, True)
    merges = get_merge_or_branch_nodes(db, False, True, True, True)
    branchings = get_merge_or_branch_nodes(db, True, False, True, False)
    del db
    unnamed_pairs = {}
    # head rule
    processed_commits = {sha: ref[0] for sha, ref in refs.items()}
    logging.info("marked heads with branch names")
    # parents of merges
    for merge_commit in merges:
        try:
            branch_names, merge_type = \
                get_parents_names(repo.commit(merge_commit), None)
            if not merge_type:
                logging.warning("No merge for %s with message %s",
                                merge_commit, repo.commit(merge_commit).message)
            if merge_type and "branch" not in str(merge_type):
                unnamed_pairs[merge_commit] = repo.commit(merge_commit).parents[0].hexsha
            else:
                for commit, name in branch_names:
                    if name:
                        processed_commits[commit.hexsha] = name
                    else:
                        # workaround for not named pairs of same branch
                        unnamed_pairs[merge_commit] = commit.hexsha
        except NoStdMergeMessageError as e:
            logging.info((merge_commit, e))
        except MessageParsingError as e:
            logging.error((merge_commit, e))
    logging.info("marked parents of merge")
    # ancestral rule
    while True:
        pre_process_copy = processed_commits.copy()
        for commit, label in pre_process_copy.items():
            commit = repo.commit(commit)
            if not commit.parents:
                continue
            commit = commit.parents[0]
            while commit.hexsha not in branchings and not processed_commits.get(commit.hexsha):
                processed_commits[commit.hexsha] = label
                if unnamed_pairs.get(commit.hexsha):
                    processed_commits[unnamed_pairs[commit.hexsha]] = label
                    del unnamed_pairs[commit.hexsha]
                commit = commit.parents[0]
        if len(processed_commits) == len(pre_process_copy):
            break
    logging.info("applied ancenstral rule")
    # majority rule
    for commit in branchings:
        commit = repo.commit(commit)
        if processed_commits.get(commit.hexsha):
            continue
        name_cntr = Counter()
        parent: git.Commit = commit
        unlabeled_commits = []
        if len(commit.parents) == 1:
            parent = commit.parents[0]
        while parent.hexsha not in merges and parent.hexsha not in branchings:
            branch = processed_commits.get(parent.hexsha)
            if branch:
                name_cntr[branch] += 1
            else:
                unlabeled_commits.append(parent.hexsha)
            if len(parent.parents) >= 1:
                parent = parent.parents[0]
            else:
                break
        if parent.hexsha in branchings:
            branch = processed_commits.get(parent.hexsha)
            if branch:
                name_cntr[branch] += 1
        if len(name_cntr) > 0:
            majority_branch = name_cntr.most_common(1)[0][0]
            processed_commits[commit.hexsha] = majority_branch
            for c in unlabeled_commits:
                processed_commits[c] = majority_branch
    return (processed_commits, edges)


# https://doi.org/10.1109/ICSME.2016.39
def get_parents_names(commit: git.Commit, current_branch_name: Optional[str] = "main(implicit)") \
        -> Tuple[List[Tuple[git.Commit, str]], Optional[str]]:
    """
    retrieves given commit's parent name from commit message

    @param commit: a commit
    @param current_branch_name: the name of the branch the commit is on
    @return: a tuple of list of (commit, branch) tuples and the merge type
    @raise MessageParsingError: the parsing ran into an error
    @raise NoStdMergeMessageError: the message does not conform to expected schema
    """
    logging.log(logging.DEBUG - 5, "Processing %s", commit.hexsha)
    parents = commit.parents
    res = []
    # ancestral rule
    if len(parents) == 1:
        return ([(parents[0], current_branch_name)], None)
    # parse commit message for branch names
    # \n split due to bad multiline Pull-Requests
    message_parts: List[str] = commit.message.lower().split("\n")[0].split()
    logging.log(1, "Decomposing %s", message_parts)
    parent_names = []
    curr_token: int = 0
    if not message_parts[curr_token] == "merge" or len(message_parts) < 2:
        raise NoStdMergeMessageError("No merge message detected: %s", commit.message)
    curr_token += 1
    merge_type = message_parts[curr_token]
    if merge_type == "pull" or merge_type.startswith("remote"):
        merge_type = message_parts[curr_token] + " " + message_parts[curr_token + 1]
        curr_token += 2
    elif merge_type in ["branch", "tag", "commit", "branches", "tags", "commits"]:
        curr_token += 1
    else:
        logging.warning("No Merge type detected: %s", commit.message)
        merge_type = "Unknown"
    octopi_merge = False
    if "and" in message_parts[curr_token:] and \
            merge_type in \
            ["tags", "commits", "branches", "remote braches", "remote-tracking branches"]:
        octopi_merge = True
    if merge_type == "pull request":
        pull_request_no = message_parts[curr_token]
        curr_token += 1
        if curr_token >= len(message_parts) or not message_parts[curr_token] == "from":
            raise MessageParsingError("pull request lacks origin description: %s", commit.message)
        curr_token += 1
    if not octopi_merge:
        parent_names = [message_parts[curr_token]]
        curr_token += 1
    else:
        while not message_parts[curr_token] == "into" or curr_token == len(message_parts):
            if not message_parts[curr_token] == "and":
                parent_names.append(message_parts[curr_token].rstrip(","))
            curr_token += 1
    if not curr_token == len(message_parts):
        # edit of original algorithm
        if message_parts[curr_token] == "from" or message_parts[curr_token] == "of":
            curr_token += 1
            branch_repo = message_parts[curr_token]
            curr_token += 1
        # end
    if not curr_token == len(message_parts):
        if message_parts[curr_token] == "into":
            curr_token += 1
            current_branch_name = message_parts[curr_token]
        else:
            logging.warning("Abitrary message postfix detected: %s", commit.message)
    del curr_token
    del message_parts
    if not len(parents) - 1 == len(parent_names):
        raise MessageParsingError(
            "NLP-detected no of parent commits does not match actual parent no."
            " Is %i should be %i",
            len(parent_names) + 1, len(parents))
    res.append((parents[0], current_branch_name))
    for i, name in enumerate(parent_names):
        if name:
            res.append((parents[i + 1], name.strip("\"").strip("\'").strip("\\")))
        else:
            logging.warning("No name for %s (parent %i) by Message",
                            parents[i + 1].hexsha, i, commit.message)
    return (res, merge_type)


class MessageParsingError(Exception):
    """
    The parsing of commit message ran into an error
    """
    pass


class NoStdMergeMessageError(MessageParsingError):
    """
    The commit message does not conform to given schema
    """
    pass



"""
Graph Properties analyzer
"""
import sqlite3 as sql
from typing import Dict, Iterable, List, NamedTuple, Optional, Tuple, Union

import git
import pandas as pd

import cocluremig.utils.gitutils as gitutils

Node = Union[str, git.Commit]
"""Graph Node"""


class Edge(NamedTuple):
    """
    Graph Edge
    """
    child: Node
    parent: Node


AdjencyMatrix = pd.DataFrame
"""Adjency matrix of graph"""
Graph = Union[List[Edge], AdjencyMatrix]
"""Graph representation"""


def edge_list_to_directed_adjency_matrix(edge_list: List[Edge], nodes: Optional[Iterable[str]] = None) -> AdjencyMatrix:
    """
    converts the given edge list to an adjency matrix

    @param edge_list: list of edges
    @param nodes: the list of nodes in the graph
    @return: the adjency matrix
    """
    shas: List[str]
    if not nodes:
        shas = set({x for e in edge_list for x in e})
    else:
        shas = set(nodes)
    shas = sorted(shas)
    res = pd.DataFrame(0, index=shas, columns=shas)
    for (child, parent) in edge_list:
        res.at[child, parent] = 1
    return AdjencyMatrix(res)


def edge_list_to_memory_db(edge_list: List[Edge]) -> sql.Connection:
    """
    Puts the edge list into an in-memory database

    @param edge_list: a graph edge list
    @return: a connection to the in-memory database
    """
    db_con = sql.connect(":memory:")
    cursor = db_con.cursor()
    cursor.execute("CREATE TABLE Edges(child TEXT,parent TEXT,annotation JSON DEFAULT NULL)")
    db_con.commit()
    for (e1, e2) in edge_list:
        cursor.execute("""INSERT INTO Edges('child', 'parent') VALUES (?,?)""", (e1, e2))
    db_con.commit()
    cursor.close()
    return db_con


def get_head_nodes(db_con: sql.Connection) -> Iterable[str]:
    """
    Gets all head nodes (no children) from given commit graph
    @param db_con: the database connection
    @return: a collection of nodes
    """
    return get_merge_or_branch_nodes(db_con, True, True, True, False)


def get_merge_nodes(db_con: sql.Connection) -> Iterable[str]:
    """
    Gets all merge nodes (at least two parents) from given commit graph
    @param db_con: the database connection
    @return: a collection of nodes
    """
    return get_merge_or_branch_nodes(db_con, False, True, True, True)


def get_branching_nodes(db_con: sql.Connection) -> Iterable[str]:
    """
    Gets all branching nodes (at least two children) from given commit graph
    @param db_con: the database connection
    @return: a collection of nodes 
    """
    return get_merge_or_branch_nodes(db_con, True, False, True, True)


def get_root_nodes(db_con: sql.Connection) -> Iterable[str]:
    """
    Gets all root nodes (no parents) from given commit graph
    @param db_con: the database connection
    @return: a collection of nodes 
    """
    get_merge_or_branch_nodes(db_con, True, True, False, True)


def get_merge_or_branch_nodes(db_con: sql.Connection,
                              ignore_merges=False,
                              ignore_branching=False,
                              ignore_root=False,
                              ignore_head=False) \
        -> Iterable[str]:
    """
    gets structural special nodes from an in-memory database containing a graph

    @param db_con: the database connection
    @param ignore_merges: ignore merge nodes (=>2 parent nodes)
    @param ignore_branching: ignore branching nodes (=>2 child nodes)
    @param ignore_root: ignore root nodes (no parent nodes)
    @param ignore_head: ignore head/leaf nodes (no child nodes)
    @return: a list of the special nodes from the graph
    """
    cursor = db_con.cursor()
    merges = []
    if not ignore_merges:
        cursor.execute("SELECT child, COUNT(parent) as no FROM Edges GROUP BY child HAVING no > 1")
        for row in cursor.fetchall():
            merges.append(row[0])
    branchings = []
    if not ignore_branching:
        cursor.execute("SELECT parent, COUNT(child) as no FROM Edges GROUP BY parent HAVING no > 1")
        for row in cursor.fetchall():
            merges.append(row[0])
    roots = []
    if not ignore_root:
        cursor.execute("SELECT parent p FROM Edges WHERE p NOT IN (SELECT child FROM Edges)")
        for row in cursor.fetchall():
            roots.append(row[0])
    heads = []
    if not ignore_head:
        cursor.execute("SELECT child p FROM Edges WHERE p NOT IN (SELECT parent FROM Edges)")
        for row in cursor.fetchall():
            heads.append(row[0])
    cursor.close()
    return set(merges + branchings + roots + heads)


def get_sequential_commits(db_con: sql.Connection) -> Iterable[str]:
    """
    gets all sequential nodes from an in-memory database containing a graph
    Sequential are all nodes with exactly one parent and exactly one child
    @param db_con: the database connection
    @return: a collection of nodes
    """
    cursor = db_con.cursor()
    cursor.execute(
        """SELECT e1.parent p FROM Edges e1 JOIN Edges e2
        ON e1.parent == e2.child GROUP BY p HAVING COUNT(*) == 1""")
    res = map(lambda x: x[0], cursor.fetchall())
    return list(res)


def get_sequential_paths(db_con: sql.Connection) -> Iterable[List[str]]:
    """
    Creates all sequential paths  from an in-memory database containing a graph

    @param db_con: the database connection
    @return: a List of paths (list of nodes)
    """
    cursor = db_con.cursor()
    # get all parts of sequential paths
    starts = []
    cursor.execute(
        """SELECT child, parent FROM Edges WHERE
        parent in
        (SELECT e1.parent p FROM Edges e1 JOIN Edges e2
        ON e1.parent == e2.child GROUP BY p HAVING COUNT(*)== 1)
        """)
    for row in cursor.fetchall():
        starts.append((row[0], row[1]))
    ends = []
    cursor.execute(
        """SELECT child, parent FROM Edges WHERE
        child in
        (SELECT e1.parent p FROM Edges e1 JOIN Edges e2
        ON e1.parent == e2.child GROUP BY p HAVING COUNT(*)== 1)""")
    for row in cursor.fetchall():
        ends.append((row[0], row[1]))
    inter = []
    cursor.execute(
        """SELECT child, parent FROM Edges WHERE
        child in
        (SELECT e1.parent p FROM Edges e1 JOIN Edges e2
        ON e1.parent == e2.child GROUP BY p HAVING COUNT(*)== 1)
        AND parent in
        (SELECT e1.parent p FROM Edges e1 JOIN Edges e2
        ON e1.parent == e2.child GROUP BY p HAVING COUNT(*)== 1)""")
    for row in cursor.fetchall():
        inter.append((row[0], row[1]))
    starts = [x for x in starts if x not in inter]
    start_finder = {r[1]: r[0] for r in starts}
    ends = [x for x in ends if x not in inter]
    end_finder = dict(ends)
    inter_starts = {x[0] for x in inter if x[0] not in map(lambda x: x[1], inter)}
    connections = dict(inter)
    # multilple sequential paths may start from the same v1
    paths = []
    for node in inter_starts:
        curr_path = [node]
        next_node = connections.get(node)
        while next_node:
            curr_path.append(next_node)
            node = next_node
            next_node = connections.get(node)
        paths.append(
            [start_finder[curr_path[0]]]
            + curr_path
            + [end_finder[curr_path[-1]]]
        )
    for sp in set(start_finder.keys()).intersection(end_finder.keys()):
        paths.append([start_finder[sp], sp, end_finder[sp]])
    # paths = set([tuple(x) for x in paths])
    return paths


def get_merge_or_branch_nodes_from_adjency_matrix(graph: Graph,
                                                  ignore_merges=False, ignore_branching=False,
                                                  ignore_root=False, ignore_head=False) \
        -> Iterable[str]:
    """
    gets structural special nodes from given graph

    @param graph: an edge-list of adjency-matrix
    @param ignore_merges: ignore merge nodes (=>2 parent nodes)
    @param ignore_branching: ignore branching nodes (=>2 child nodes)
    @param ignore_root: ignore root nodes (no parent nodes)
    @param ignore_head: ignore head/leaf nodes (no child nodes)
    @return: a list of the special nodes from the graph
    """
    adjency_matrix: AdjencyMatrix
    if not isinstance(graph, AdjencyMatrix):
        # transform to adjency_matrix
        adjency_matrix = edge_list_to_directed_adjency_matrix(graph)
    else:
        adjency_matrix = graph
    (m_nodes_in, m_nodes_root, m_nodes_out, m_nodes_head) = ([], [], [], [])
    # get number of in/out-going edges
    (edge_count_out, edge_count_in) = (None, None)
    if not (ignore_merges and ignore_root):
        edge_count_out = adjency_matrix.sum()
    if not ignore_merges:
        m_nodes_out = edge_count_out.index[edge_count_out > 1].tolist()
    if not ignore_root:
        m_nodes_root = edge_count_out.index[edge_count_out == 0].tolist()
    if not (ignore_branching and ignore_head):
        edge_count_in = adjency_matrix.sum(1)
    if not ignore_branching:
        m_nodes_in = edge_count_in.index[edge_count_in > 1].tolist()
    if not ignore_head:
        m_nodes_head = edge_count_in.index[edge_count_in == 0].tolist()
    return set(m_nodes_in + m_nodes_root + m_nodes_out + m_nodes_head)


def get_head_refs(repo: git.Repo, db_con: Optional[sql.Connection] = None) -> Dict[str, str]:
    """
    Gets the names of graph's head nodes

    @param repo: a git repo
    @param db_con: the edge list db (generated if not set)
    @return: a dict of sha to name
    """
    if not db_con:
        db_con = edge_list_to_memory_db(gitutils.get_edge_list(repo))
    return {k: v
            for k, v in gitutils.get_ref_names(repo).values()
            if k in
            get_merge_or_branch_nodes
            (db_con, True, True, True, False)}


def adjency_matrix_to_edge_list(adjency_matrix: AdjencyMatrix) -> Tuple[str, str, int]:
    """
    Converts the adjency- matrix to an edge list with weight

    @param adjency_matrix: the matrix to convert
    @return: an edge list with weight
    """
    # https://stackoverflow.com/questions/48218455/how-to-create-an-edge-list-dataframe-from-a-adjacency-matrix-in-python
    adjency_matrix = adjency_matrix.rename_axis('Child') \
        .reset_index() \
        .melt('Child', value_name='Weight', var_name='Parent') \
        .query('Child != Parent') \
        .reset_index(drop=True)
    adjency_matrix = adjency_matrix[adjency_matrix['Weight'] != 0]
    return adjency_matrix.values.tolist()



"""
Aggregators for file mining results
"""
from collections import Counter
from typing import Dict, \
    Union

from cocluremig.analyzer.commit.util import AverageElement


def additive_dict_merge(dict_1: Dict[str, Union[int, float, AverageElement]],
                        dict_2: Dict[str, Union[int, float, AverageElement]]) \
        -> Dict[str, Union[int, float, AverageElement]]:
    """
    Dictionary per key value addition
    @param dict_1: a dictionary
    @param dict_2: another dictionary
    @return: A dictionary containing added values
    """
    dict_1 = Counter(dict_1)
    dict_1.update(dict_2)
    return dict_1



# pylint: disable=line-too-long, bad-indentation, trailing-whitespace

from typing import Optional, TextIO, Tuple

from cocluremig.visualization.dot_base_writer import BaseDotWriter


class PrettyDotWriter(BaseDotWriter):
    """
    Human readable dot file with indents and line-breaks
    """

    def __init__(self, file_writer: TextIO):
        BaseDotWriter.__init__(self, file_writer)

    def write_head(self, name: str):
        BaseDotWriter.write_head(self, name)
        self._file_writer.write("\n")

    def write_edge(self, edge: Tuple[str, str]):
        self._file_writer.write("\t")
        BaseDotWriter.write_edge(self, edge)
        self._file_writer.write("\n")

    def write_vertex(self, vertex: str, label: Optional[str] = None, color: Optional[str] = None):
        self._file_writer.write("\t")
        BaseDotWriter.write_vertex(self, vertex, label, color)
        self._file_writer.write("\n")



"""
All functions working directly on git blob objects
"""

import logging
import sys
from typing import Any, Dict, \
    Optional

import git
import lizard
import magic

from cocluremig.analyzer.commit.util import AVERAGE_NEUTRAL, AverageElement


def get_code_metrics_lizard(sha: git.objects.Blob) -> Dict[str, Any]:
    """
    applies lizard analyzer (http://lizard.ws) onto blob

    @param sha: the blob
    @return: lizard results
    """
    text = get_text_representation(sha)
    if text:
        return lizard.analyze_file.analyze_source_code(sha.path, text).__dict__
    return {}


def get_lizard_metrics_avarageable(sha: git.objects.Blob) -> Dict[str, AverageElement]:
    """
    applies lizard onto the blob and comutes average (mean) values of the results

    @param sha: the blob
    @return: average results per value
    """
    plain_lizard = get_code_metrics_lizard(sha)
    # {'nloc': 9, 'function_list': [<lizard.FunctionInfo object at 0x10bf7af10>],
    # 'filename': '../cpputest/tests/AllTests.cpp'}
    res = {}
    if not plain_lizard:
        return res
    res['nloc_file'] = AverageElement(plain_lizard['nloc'])
    if plain_lizard['function_list']:
        for function in plain_lizard['function_list']:
            function = function.__dict__
            # {'cyclomatic_complexity': 1, 'token_count': 22,
            # 'name': 'main', 'parameter_count': 2, 'nloc': 3,
            # 'long_name': 'main( int ac , const char ** av )',
            # 'start_line': 30}
            for key in ['cyclomatic_complexity',
                        'token_count', 'top_nesting_level',
                        'fan_in', 'fan_out', 'nloc']:
                res[key] = res.get(key, AVERAGE_NEUTRAL) \
                           + AverageElement(function[key])
            res['parameter_count'] = \
                res.get('parameter_count', AVERAGE_NEUTRAL) \
                + AverageElement(len('full_parameters'))
        res['nloc_function'] = res.pop('nloc')
    return res


def get_file_extension(sha: git.objects.Blob) -> str:
    """
    get file extension for blob in tree

    @param sha: the blob
    @return: a file extension
    """
    return str(sha.path).split("/")[-1].split(".")[-1]


def get_lines_of_text(sha: git.objects.Blob) -> Optional[int]:
    """
    Computes lines of text from blob
    Returns None if blob is not decodeable as text

    @param sha: the blob
    @return: the lines of text
    """
    text = get_text_representation(sha)
    if text is not None:
        return text.count("\n") + 1
    return None


def get_object_size(sha: git.objects.Blob) -> Optional[int]:
    """
    Gets blob size

    @param sha: the blob
    @return: the object size
    """
    return sys.getsizeof(sha.data_stream.read())


def _get_mixed_size_per_extension(sha: git.objects.Blob) -> Dict[str, AverageElement]:
    ext = get_file_extension(sha)
    size = get_lines_of_text(sha)
    if not size:
        size = get_object_size(sha)
    return {ext: AverageElement(size)}


def get_text_representation(sha: git.objects.Blob) -> Optional[str]:
    """
    decodes the blob into text
    returns None if object is not decodable
    decoding failures are printed with logging.error
    decoding problems (broken encoding) are printed with logging.warn and return text anyway

    @param sha: the object sha
    @return: the text
    """
    enc = _get_encoding(sha)
    try:
        # only decode non-binaries e.g."binary", "application/tarbinary"
        if enc.find("binary") < 0:
            return str(sha.data_stream.read().decode(enc))
    except LookupError as error:
        logging.error("Unable to decode Object")
        logging.error((error, sha.path, sha.hexsha))
        if enc in __ENCODING_REPLACEMENTS.keys():
            return str(sha.data_stream.read()
                       .decode(__ENCODING_REPLACEMENTS[enc],
                               errors="replace"))
        raise error
    except UnicodeDecodeError as error:
        logging.warning("Malformed Unicode BLObject")
        logging.warning((error, sha.path, sha.hexsha))
        return str(sha.data_stream.read().decode(enc, errors="replace"))
    return None


def _get_encoding(sha: git.objects.Blob) -> str:
    magician = magic.Magic(mime_encoding=True)
    return magician.from_buffer(sha.data_stream.read())


__ENCODING_REPLACEMENTS = {'unknown-8bit': 'utf-8'}



"""
Basic utilites
"""

import json


class JSONEncoder(json.JSONEncoder):
    """
    JSONEncoder which is more robust against some input data
    - replaces all iterables with lists
    """

    def default(self, o):
        try:
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)



"""
Runtime tweaks
"""
import psutil


def get_max_threads() -> int:
    """
    gets number of cpu threads
    @return: cpu thread count
    """
    return psutil.cpu_count()


def get_cache_size() -> int:
    """
    gets a cache size depending on available and total memory
    @return: cache size
    """
    # get free size in KB
    total = int(psutil.virtual_memory().total / 1024)
    avail = int(psutil.virtual_memory().available / 1024)
    return __nearest_power_of_two(max([1024 * 1024, int(total / 4), int(avail / 2)]))


# if you want a power of 2 by force
def __nearest_power_of_two(number: int) -> int:
    """
    Gets the nearest power of 2 from given number

    @param number: a number
    @return: a power of 2
    """
    exp = number.bit_length()
    # next greater power of 2
    greater_value = 1 << exp
    # next smaller power of 2
    smaller_value = 1 << (exp - 1)
    if abs(greater_value - number) > abs(smaller_value - number):
        return smaller_value
    return greater_value



"""
Graph Statisic Getters
"""
import math
import statistics
from numbers import Real
from typing import List, NamedTuple

from cocluremig.analyzer import graph_analyzer


class GraphMetrics(NamedTuple):
    """
    Graph Metrics Container
    """
    edge_count: int
    """Number of edges"""

    node_count: int
    """Number of Nodes/Vertices"""

    backwards_edge_count: int = 0
    """Number of direct backwards count"""


def get_graph_metrics(graph: graph_analyzer.Graph) -> GraphMetrics:
    """
    calculates basic metrics for given graph

    @param graph: the graph to analyze
    @return: metrics
    """
    if isinstance(graph, graph_analyzer.AdjencyMatrix):
        graph = [(x[0], x[1]) for x in graph_analyzer.adjency_matrix_to_edge_list(graph)]
    edge_count = len(graph)
    node_count = len({node for edge in graph for node in edge})
    backwards_edge_count = int(len({(n1, n2) for (n1, n2) in graph if (n2, n1) in graph}) / 2)
    return GraphMetrics(edge_count, node_count, backwards_edge_count)


class NodeTypeCount(NamedTuple):
    """
    Numbers of nodes per type
    """
    head_count: int
    """Number of heads (commits without children)"""
    root_count: int
    """Number of roots (commits without parents)"""
    branching_count: int
    """Number of branchings (commits with at least two children)"""
    merge_count: int
    """Number of merges (commits with at least two parents)"""
    sequential_count: int
    """Number of sequential commits (exactly one parent and one child)"""


def get_node_types_count(graph: graph_analyzer.Graph) -> NodeTypeCount:
    """
    Calaculates node counts per type

    @param graph: the graph to analyze
    @return: metrics
    """
    if isinstance(graph, graph_analyzer.AdjencyMatrix):
        graph = [(x[0], x[1]) for x in graph_analyzer.adjency_matrix_to_edge_list(graph)]
    db_con = graph_analyzer.edge_list_to_memory_db(graph)
    head_count = len(graph_analyzer.get_merge_or_branch_nodes(db_con, True, True, True, False))
    root_count = len(graph_analyzer.get_merge_or_branch_nodes(db_con, True, True, False, True))
    branching_count = len(graph_analyzer.get_merge_or_branch_nodes(db_con, True, False, True, True))
    merge_count = len(graph_analyzer.get_merge_or_branch_nodes(db_con, False, True, True, True))
    sequential_count = len(graph_analyzer.get_sequential_commits(db_con))
    return NodeTypeCount(head_count, root_count, branching_count, merge_count, sequential_count)


class ClusterStats(NamedTuple):
    """
    A simple Container for basic cluster analysis
    """
    cluster_cnt: int
    """Number of clusters"""

    clustered_nodes_total: int
    """Number of nodes in clusters"""

    cluster_max_size: int
    """Size of biggest cluster"""

    cluster_min_size: int
    """Size of smallest cluster"""

    cluster_median_size: Real
    """Median size of clusters"""

    cluster_avg_size: float
    """Average (mean) size of clusters"""

    cluster_size_variance: float
    """Variance of cluster size"""


def calculate_cluster_stats(clusters: List[List[str]]) -> ClusterStats:
    """
    Calculates some basic statistics about given clusters

    @param clusters: a list of clusters (as list of elements)
    @return: the statistics
    """
    cluster_sizes: List[int] = list(map(len, clusters))
    total_clustered_nodes = sum(cluster_sizes)
    no = len(cluster_sizes)
    maxs = max(cluster_sizes)
    mins = min(cluster_sizes)
    medians = statistics.median(cluster_sizes)
    avg_s = statistics.fmean(cluster_sizes)
    variance = math.nan if no < 2 else statistics.variance(cluster_sizes, avg_s)
    return ClusterStats(no, total_clustered_nodes, maxs, mins, medians, avg_s, variance)



import os
from typing import Iterable, Optional, TextIO, Tuple

from cocluremig.utils.gitutils import normalize_id_name
from cocluremig.visualization.dot_writer import BaseDotWriter
from pynpmd import JsLibResolver


class DotHtmlRenderWriter(BaseDotWriter):
    """
    html-based dot-based view using d3-graphviz
    """

    def __init__(self, file_writer: TextIO):
        BaseDotWriter.__init__(self, file_writer)
        self._libs: Iterable[str] = []

    # noinspection PyArgumentEqualDefault
    def __dep_resolve(self):
        modules = ["jquery", "d3", "@hpcc-js/wasm", "d3-graphviz"]
        rslvr = JsLibResolver("../js_npm_lib_cache", True)
        for module in modules:
            path = rslvr.get_lib(module)
            self._libs += [path]
        # store abspaths on runtime
        self._libs = map(os.path.abspath, self._libs)

    def _write_js_deps(self, inline=False):
        fw = self._file_writer
        if not self._libs:
            self.__dep_resolve()
        for lib in self._libs:
            if inline:
                fw.write("<script type='application/javascript'>")
                with open(lib, "r") as reader:
                    fw.write(reader.read())
            else:
                fw.write("<script src=\"")
                fw.write(os.path.relpath(lib, os.path.dirname(fw.name)))
                fw.write("\">")
            fw.write("</script>\n")

    def write_head(self, name: str):
        fw = self._file_writer
        fw.write("<!DOCTYPE html>\n")
        fw.write("<head>\n")
        fw.write("<meta charset=\"utf-8\">\n")
        # fw.write("<meta http-equiv=\"Content-Security-Policy\"
        # content=\"default-src *; style-src 'self'  "
        #         "'unsafe-inline'; script-src 'self' 'unsafe-inline' 'unsafe-eval'\">")
        fw.write("<title>")
        fw.write(normalize_id_name(name))
        fw.write("</title>\n")
        self._write_js_deps()
        fw.write("</head>\n")
        fw.write("<body>\n")

    def write_simple_render_head(self, name: str):
        """
        Write render init

        @param name: the graphs id
        """
        fw = self._file_writer
        fw.write("<div id=\"graph\" style=\"text-align: center;\"></div>\n")
        fw.write("<script>\n")
        fw.write("head = '")
        BaseDotWriter.write_head(self, name)
        fw.write("';\n")
        fw.write("data = [''+\n")

    def write_edge(self, edge: Tuple[str, str]):
        fw = self._file_writer
        fw.write("\t'")
        BaseDotWriter.write_edge(self, edge)
        fw.write("'+\n")

    def write_vertex(self, vertex: str, label: Optional[str] = None, color: Optional[str] = None):
        fw = self._file_writer
        fw.write("\t'")
        BaseDotWriter.write_vertex(self, vertex, label, color)
        fw.write("'+\n")

    def write_simple_render_foot(self):
        """
        Write render end
        """
        fw = self._file_writer
        fw.write("\t''];\n")
        fw.write("cdata = '';\n")
        fw.write("for (i=0; i<data.length; i++){\n")
        fw.write("\t")
        fw.write("cdata=cdata+data[i];\n")
        fw.write("\t")
        fw.write("console.log(i);\n\tconsole.log(data[i]);\n")
        fw.write("\t")
        fw.write("d3.select(\"#graph\").graphviz().renderDot(")
        fw.write("head+cdata+'}'")
        fw.write(");\n")
        # fw.write("}")
        # fw.write("');")
        fw.write("}\n")
        fw.write("</script>\n")

    def write_footer(self):
        fw = self._file_writer
        fw.write("</body>")
